from scipy.special import comb
from sklearn.metrics import roc_auc_score
import numpy as np

class AUC:

    def __init__(self, multi_target=False):
        self.name = "stochastic_auc"

    
    def __call__(self, true_target, pred_target):
        pass


import numpy as np
import tensorflow as tf
from tensorflow.keras.metrics import get


class Accuracy:

    def __init__(self, name="accuracy"):
        self.name = name
        self.__fn = get(name)

    
    def __call__(self, true_targets, predictions, **kwargs):
        exp, _var = predictions
        true_targets = tf.convert_to_tensor(true_targets)
        output = self.__fn(true_targets, exp)
        return np.mean(output)



import numpy as np
import tensorflow as tf


class Loss:
    """
        Calculate the loss using a predefined tensorflow function.
    """

    def __init__(self, name="loss"):
        self.name = name

    def __call__(self, true_target, predictions, **kwargs):
        """
            Parameters:
                true_target (numpy.ndarray): The true targets to check against.
                pred_target (numpy.ndarray): The targets predicted by the network. Shape (num datapoints, num samples, ...)
            
            Returns:
                (int) the loss calculated.
        """

        # if sample_size is None:
        #     raise ValueError("Error in tf_al.stats.sampling.Loss(). Missing kwarg parameter sample_size. Seems like you arent using a McDropout model.")

        pred_mean = np.mean(predictions, axis=1)
        loss_fn = tf.keras.metrics.get(self.name)
        return np.mean(loss_fn(true_target, pred_mean).numpy())


import numpy as np

class Confusion:

    def __init__(self, multi_class=True, labels=None):
        self.name = "true_positive"
        self._multi_class = multi_class
        self._labels = labels

    
    def __call__(self, true_targets, predictions, **kwargs):
        """
        
        """


        num_classes = predictions.shape[-1]
        class_labels = self._get_labels(num_classes)
        
        true_positive_scores = np.zeros(num_classes)
        for idx in range(num_classes):
            label = class_labels[idx]

            selector = true_targets == label
            
        


        # Binary case
        # predictions = np.where(predictions>.5, 1, 0)
        # stacked_targets = np.vstack([true_targets]*num_samples).T
        # predicted_labels = np.argmax(predictions, axis=-1)
        # return np.sum(stacked_targets == predicted_labels]

    
    def _get_labels(self, num_classes):
        if self._labels is not None:
            return self._labels

        return list(range(num_classes))




def prepend(prefix, name):
    if prefix is not None:
        return prefix + "_" + name
    
    return name


def is_loss_metric(metric_name):
    return "entropy" in metric_name \
            or "loss" in metric_name \
            or "hinge" in metric_name \
            or "error" in metric_name
        
   

def __get_stochastic_metric(metric_name):
    from .stochastic import Loss, Accuracy
    
    metric_name = metric_name.replace("stochastic_", "")
    # if metric_name == "auc":
    #     return AUC

    if "accuracy" in metric_name:
        return Accuracy(metric_name)
    
    elif is_loss_metric(metric_name):
        return Loss(metric_name)
    
    raise ValueError("Error in stats.get(). Couldn't find any metric for {}.".format(metric_name))


def __get_sampling_metric(metric_name):
    from .sampling import AUC, TruePositive, Loss, Accuracy

    metric_name = metric_name.replace("sampling_", "")
    if metric_name == "auc":
        return AUC
    
    elif "accuracy" in metric_name:
        return Accuracy(metric_name)
    
    elif is_loss_metric(metric_name):
        return Loss(metric_name)

    raise ValueError("Error in stats.get(). Couldn't find any metric for {}.".format(metric_name))
    

def get(prefix, metric_name):
    """
        Transform a metric if possible into a
        custom model metric used.
        
        If no metric was found try a fallback on regular tensorflow metric methods.
    """
    
    metric_name = prepend(prefix, metric_name)
    if "stochastic_" in metric_name:
        return __get_stochastic_metric(metric_name)

    if "sampling_" in metric_name:
        return __get_sampling_metric(metric_name)






class FalsePositive:

    def __init__(self, multi_label=False):
        self.name = "false_positive"
        self._multi_label = multi_label

    
    def __call__(self, true_targets, predictions, **kwargs):
        pass


from .logger import setup_logger
from .pools import init_pools
from .tf import setup_growth, set_tf_log_level, disable_tf_logs
from .problem import ProblemUtils
from .acquisition import beta_approximated_upper_joint_entropy


from types import FunctionType



class BaseMetric:
    """
        A simple metric object. Encapsulating the callback function
        for a specific custom metric as well as the prefix to be applied to the 
        returned metric keys.
    """

    def __init__(self, callback: FunctionType, prefix: str):

        if not isinstance(callback, FunctionType) and not hasattr(callback, "__call__"):
            raise ValueError("Error in BaseMetric.__init__(callback, prefix). Expected callback to be a function.")

        self.__callback = callback
        self.prefix = prefix

    
    def __call__(self, *args) -> dict:
        return self.__callback(*args)



class ModelSpecificMetric:
    """
        Specify metrics that will be applied only to specific types of model types.

        Parameters:
            prefix (str): a prefix that will prepended to the metric returned by the callback.
    """

    def __init__(self, prefix: str=None):
        self.prefix = prefix
        self.__callbacks = {}

    
    def __call__(self, model_type: str, *args):
        """
            Applies a callback function on the arguments when registered for given
            model type. 

            Following parameters will be passed down in order as args to the callback.

            Parameters:
                    prediction (numpy.ndarray): The predictions made by the model.
                    inputs (numpy.ndarray): The inputs values to the network.
                    true_targets (numpy.ndarray): The true target values of the dataset.
        """
        if model_type in self.__callbacks.keys():
            return self.__callbacks[model_type](*args)

        return  {}
    

    def add_callback(self, model_type: str, callback: FunctionType):
        """
            Register callback methods to be applied to models of a specific type.
        
            Parameters:
                model_type (str): The model type the callback function will be applied on.
                callback (function): 
        """

        if not isinstance(callback, FunctionType):
            raise ValueError("Error in ModelSpecificMetric.add_metric(). Expected parameter callback to be a function.")

        self.__callbacks[model_type] = callback



class MetricsAccumulator:
    """
        Collect metrics from the active learning loop to be added to
        the output per iteration.

        Allows definition of custom callback functions to be applied to
        the evaluation step of the active learning loop.

        Predefined metrics that will be collected are:
            - train_time: The time it took to train the model.
            - optim_time: The time used to execute the optimization
            - eval_time: The time it took to perform the evaluation step 

        Parameters:
            metrics (str|list(str)): A name or a list of predefined metric names. (default=None) 
    """

    def __init__(self):
        self.__metrics = []

        
    def __call__(self, *args, **kwargs) -> dict:
        """
            Applies the callback functions and passes the
            parameters down as kwargs.

            Following parameters are getting passed down to the metric callback as 
            in order as args.

            Parameters:
                model_type (str): The model type in lower case letters. 
                prediction (numpy.ndarray): The predictions made by the model.
                inputs (numpy.ndarray): The inputs values to the network.
                true_targets (numpy.ndarray): The true target values of the dataset.

            Returns:
                (dict) all metrics accumulated
        """
        all_metrics = {}
        for metric in self.__metrics:
            outputs = metric(*args, **kwargs)
            outputs = self.__add_prefix(metric.prefix, outputs)
            all_metrics.update(outputs)

        return all_metrics

    
    def track(self, callback: FunctionType) -> None:
        """
            Adds a callback method for metric accumulation. The callback function
            should receive a dictionary of values and return a dictionary with unique keys.

            Pre-defined keys that should not be used are:

                - train_loss
                - train_<accuracy_name>
                - train_time
                - query_time

            Parameters:
                callback (function|BaseMetric|ModelSpecificMetric): A function return a dictionary of metrics.
        """

        if isinstance(callback, BaseMetric):
            self.__metrics.append(callback)
            return
        
        elif isinstance(callback, ModelSpecificMetric):
            self.__metrics.append(callback)
            return

        elif not isinstance(callback, FunctionType) and not hasattr(callback, "__call__"):
            raise ValueError("Error in MetricAccumulator.add_metric(). Expected parameter callback to be of type BaseMetric, ModelSpecificMetric or function. Got {}".format(type(callback)))


        # Function passed
        new_metric = BaseMetric(callback, None)
        self.__metrics.append(new_metric)


    # def add_filter(self, names):
    #     """
    #         Pre-filter metrics to outputed per active learning iteration.

    #         Parameters:
    #             names (str|list(str)): A single metric name or list of names to be remove from the iteration output.
    #     """
    #     pass



    # --------------
    # Utilties
    # -------------------

    def __add_prefix(self, prefix: str, outputs: dict) -> dict:
        """
            Adds the given prefix to the metric names.

            Parameters:
                prefix (str): A string to be prepended to the metric names.

            Returns:
                (dict) output of previous metric callback with added prefix to the metric keys.
        """

        if prefix is None:
            return outputs

        new_outputs = {}
        for key, value in outputs.items():
            new_key = prefix + "_" + key
            new_outputs[new_key] = value
        
        return new_outputs




class FalseNegative:

    def __init__(self, multi_label=False):
        self.name = "false_negative"
        self._multi_label = multi_label

    
    def __call__(self, true_targets, predictions, sample_size=None, **kwargs):
        pass


import numpy as np
from scipy.special import beta, digamma

def beta_approximated_upper_joint_entropy(a, b):
    """
        
    """
    regular_beta = beta(a, b)

    regular_beta_clipped = regular_beta*(regular_beta>0)
    # is_inf = np.any(np.isposinf(regular_beta))
    # print("Infininte",is_inf)
    # print("NaN: ", np.any(np.isnan(regular_beta)))
    # print("Negatives: ",np.any(regular_beta_clipped<0))
    # print("---------")

    sub_1 = (a/(a+b))*np.log(regular_beta_clipped, where=regular_beta_clipped!=0)
    sub_2 = (np.divide(beta(a+1, b), regular_beta_clipped, where=regular_beta_clipped!=0))

    dig_1 = digamma(a+1)
    dig_2 = digamma(a+b+1)
    dig_3 = digamma(b)

    return np.sum(sub_1-(a*sub_2*(dig_1-dig_2))-((b-1)*sub_2*(dig_3-dig_2)), axis=-1)


import math
import numpy as np
from sklearn.metrics import roc_auc_score
from scipy.special import digamma, beta
import tensorflow.keras as keras
import tensorflow as tf

from . import  Model
from ..utils import beta_approximated_upper_joint_entropy



class McDropout(Model):
    """
        Wrapper class for neural networks.


    """

    def __init__(self, model, config=None, **kwargs):
        super().__init__(model, config=config, model_type="mc_dropout", **kwargs)
        self._approximation_type = "sampling"

        # disable batch norm
        # super().disable_batch_norm()


    def __call__(self, inputs, sample_size=10, batch_size=None, **kwargs):
        """
            Perform a prediction using mc dropout as bayesian approximation.

            Parameters:
                inputs (numpy.ndarray): Inputs going into the model
                sample_size (int): Number of samples to acquire from bayesian model. (default=10)
                batch_size (int): In how many batches to split the data? (default=None)
        """

        if batch_size is None:
            batch_size = len(inputs)
        
        if batch_size < 1:
            raise ValueError("Error in McDropout.__call__(). Can't select negative amount of batches.")

        if sample_size < 1:
            raise ValueError("Error in McDropout.__call__(). Can't sample negative amount of times.")

        total_len = len(inputs)
        num_batches = math.ceil(total_len/batch_size)
        batches = np.array_split(inputs, num_batches, axis=0)
        predictions = []

        for batch in batches:

            # Sample n_times for given batch
            posterior_samples = []
            for i in range(sample_size):
                posterior_samples.append(self._model(batch, training=True))
                
            # Sampled single time or multiple times?
            if sample_size > 1:
                stacked = np.stack(posterior_samples, axis=1)
                predictions.append(stacked)
            else:
                predictions.append(posterior_samples[0])

        if len(predictions) == 1:
            return predictions[0]

        return np.vstack(predictions)



    def evaluate(self, inputs, targets, sample_size=10, **kwargs):
        """
            Evaluate a model on given input data and targets.
        """
        
        if len(inputs) != len(targets):
            raise ValueError("Error in McDropout.evaluate(). Targets and inputs not of equal length.")

        # Returns: (batch_size, sample_size, target_len) or (batch_size, target_len)
        predictions = self.__call__(inputs, sample_size=sample_size, **kwargs)

        output_metrics = {}
        for metric in self.eval_metrics:

            metric_name = None
            if hasattr(metric, "__name__"):
                metric_name = metric.__name__

            else:
                metric_name = metric.name

            output_metrics[metric_name] = metric(targets, predictions, sample_size=sample_size)

        return output_metrics

        # if self.is_classification():
        #     loss, acc = self.__evaluate(predictions, targets, sample_size)
        #     return {"loss": loss, "accuracy": acc}

        # loss_fn = keras.losses.get(self._model.loss)
        # loss = loss_fn(predictions, targets).numpy()
        # return {"loss": np.mean(loss, axis=-1), "accuracy": []}
    

    def __evaluate(self, predictions, targets, sample_size):
        """
            Parameters:
                predictions (numpy.ndarray): The predictions made by the network of shape (batch, targets) or (batch, samples, targets)
                targets (numpy.ndarray): The target values
                sample_size (int): The number of samples taken from posterior.

            Returns:
                (list()) of values representing the accuracy and loss
        """
        
        expectation = predictions
        if len(predictions.shape) == 3:
            expectation = np.average(predictions, axis=1)

        # Will fail in regression case!!!! Add flag to function?
        loss_fn = tf.keras.losses.get(self._model.loss)
        loss = loss_fn(targets, expectation)

        # Extend dimension in binary case 
        extended = self._problem.extend_binary_predictions(predictions)
        pred_targets = np.argmax(extended, axis=-1)

        # One-hot vector passed
        if len(targets.shape) == 2:
            targets = np.argmax(targets, axis=1)
        
        # Extend target dimension (multiple sample in prediction)
        if sample_size > 1:
            targets = np.vstack([targets]*sample_size).T
        
        acc = np.mean(pred_targets == targets)
        return [np.mean(loss.numpy()), acc]


    def compile(self, *args, **kwargs):
        self._model.compile(**kwargs)
        metrics = self._create_init_metrics(kwargs)        
        metric_names = self._extract_metric_names(metrics)

        self.eval_metrics = self._init_metrics("sampling", metric_names)
        self._compile_params = kwargs


    # -----
    # Acquisition functions
    # ---------------------------

    def get_query_fn(self, name):

        fn = None
        if name == "max_entropy":
            fn = self.__max_entropy
        
        elif name == "bald":
            fn = self.__bald
        
        elif name == "max_var_ratio":
            fn = self.__max_var_ratio

        elif name == "std_mean":
            fn = self.__std_mean

        elif name == "baba":
            fn = self.__baba

        return fn


    def __max_entropy(self, data, sample_size=10, **kwargs):
        """
            Select datapoints by using max entropy.

            Parameters:
                model (tf.Model) The tensorflow model to use for selection of datapoints
                unlabeled_pool (Pool) The pool of unlabeled data to select
        """        
        # Create predictions
        predictions = self.__call__(data, sample_size=sample_size)
        expectation = self.expectation(predictions)
        
        # Absolute value to prevent nan values and + 0.001 to prevent infinity values
        log_post = np.log(np.abs(expectation) + .001)

        # Calculate max-entropy
        return -np.sum(expectation*log_post, axis=1)


    def __bald(self, data, sample_size=10, **kwargs):
        # TODO: dimensions do not line up in mutli class
        # predictions shape (batch, num_predictions, num_classes)
        predictions = self.__call__(data, sample_size=sample_size)
        posterior = self.expectation(predictions)

        entropy = -self.__shannon_entropy(posterior)
        # first_term = -np.sum(posterior*np.log(np.abs(posterior) + .001), axis=1)

        # Missing dimension in binary case?
        predictions = self._problem.extend_binary_predictions(predictions)
        inner_sum = self.__shannon_entropy(predictions)
        # inner_sum = np.sum(predictions*np.log(np.abs(predictions) + .001), axis=1)
        disagreement = np.sum(inner_sum, axis=1)/predictions.shape[1]
        return entropy + disagreement


    def __max_var_ratio(self, data, sample_size=10, **kwargs):
        """
            Select datapoints by maximising variation ratios.

            # (batch, predictions, classes) reduce to (batch, predictions (max-class))
            # 1 - (count of most common class / num predictions)
        """
        predictions = self.__call__(data, sample_size=sample_size)
        posterior = self.expectation(predictions)

        # Calcualte max variation rations
        return 1 - posterior.max(axis=1)


    def __std_mean(self, data, sample_size=10,  **kwargs):
        """
           Maximise mean standard deviation.
           Check std mean calculation. Depending the model type calculation of p(y=c|x, w) can differ.
           (Kampffmeyer et al. 2016; Kendall et al. 2015)

           Todo:
            Implement distinction for different model types.
        """
        # TODO: generalize for n-classes For binary classes
        predictions = self.__call__(data, sample_size=sample_size)

        # Calculate variance/standard deviation from samples
        variance = self.variance(predictions)
        std = np.square(variance)

        # Mean over target variables
        return np.mean(std, axis=-1)


    def __baba(self, data, sample_size=10, **kwargs):
        """
            Normalized mutual information

            Implementation of acquisition function described in:
            BABA: Beta Approximation for Bayesian Active Learning, Jae Oh Woo
        """
        # predictions shape (batch, num_predictions, num_classes)
        predictions = self.__call__(data, sample_size=sample_size)
        sample_mean = self.expectation(predictions)
        entropy = -self.__shannon_entropy(sample_mean)
        disagreement = self.__disagreement(predictions)
        bald_term = self.__mutual_information(entropy, disagreement)
        
        # Beta approximation parameters
        sample_var = self.variance(predictions)
        a = ((np.power(sample_mean, 2)*(1-sample_mean))/(sample_var+.0001))-sample_mean
        b = ((1/sample_mean)-1)*a
        upper_joint_entropy = beta_approximated_upper_joint_entropy(a, b)
        return bald_term/np.abs(upper_joint_entropy)



    # --------------
    # Utils
    # --------------------

    def __disagreement(self, predictions):
        predictions = self._problem.extend_binary_predictions(predictions)
        inner_sum = self.__shannon_entropy(predictions)
        return np.sum(inner_sum, axis=1)/predictions.shape[1]


    def __mutual_information(self, entropy, disagreement):
        return entropy + disagreement

    def __shannon_entropy(self, values):
        """
            Calculate the shannon entropy for given values.
        """
        return np.sum(values*np.log(values + .001), axis=1)

    
    def expectation(self, predictions):
        """
            Calculate the mean of the distribution
            output distribution.

            Returns:
                (numpy.ndarray) The expectation per datapoint
        """
        # predictions -> (batch_size, num_predictions)
        predictions = self._problem.extend_binary_predictions(predictions)
        return np.average(predictions, axis=1)


    def variance(self, predictions):
        """
            Calculate the variance of the distribution.

            Returns:
                (numpy.ndarray) The variance per datapoint and target
        """
        predictions = self._problem.extend_binary_predictions(predictions)
        return np.var(predictions, axis=1)


    def std(self, predictions):
        """
            Calculate the standard deviation.

            Returns:
                (numpy.ndarray) The standard deviation per datapoint and target
        """
        predictions = self._problem.extend_binary_predictions(predictions)
        return np.std(predictions, axis=1)



import numpy as np
from .__checks import has_same_len, has_shape_len, of_type


def __multi_experiment_leff(main: np.ndarray, baseline: np.ndarray) -> tuple:
    """
        Calcualte labeling efficiency for multi experiment inputs.

        Parameters:
            main (numpy.ndarray): Accuracy of different experiments run.
            baseline (numpy.ndarray): Accuracy of different experimental runs of the baseline function.

        Returns:
            (tuple(float, float)) the mean of the labeling efficiency and standard deviation.
    """
    
    # Do inputs have the same number of accuracies at round x
    if main.shape[-1] != baseline.shape[-1]:
        raise ValueError("Error in leff(). Input shape missmatch. Got shapes \"main={}\" and \"baseline={}\"".format(main.shape, baseline.shape))

    main_eff = main/baseline
    return np.mean(main_eff, axis=0), np.std(main_eff, axis=0)

    # main_mean = np.mean(main, axis=0)
    # main_std = np.std(main, axis=0)
    
    # base_mean = np.mean(baseline, axis=0)
    # base_std = np.std(baseline, axis=0)

    # return (main_mean/base_mean), (main_std/base_std)


def leff(main, baseline):
    """
        Calculate the labeling efficiency over n-active learning rounds.

        Parameters:
            main (numpy.ndarray|float|int): The accuracies of main acquisition function for which to compute relative values.
            baseline (numpy.ndarray|float|int): The accuraciesk of the baseline acquisition function.

        Returns:
            (numpy.ndarray|float|int|tuple) the computed labeling efficiency per active learning round. Alternativly mean and std of computed labeling efficiency on pasesd 2D numpy array.
    """


    if of_type(np.ndarray, main, baseline):

        if not has_same_len(main, baseline):
            raise ValueError("Error in qeff(). Expected length of input parameters to match. Received \"main={}\" and \"baseline={}\".".format(main.shape, baseline.shape))

        # Output of single experiment passed (accuracy_at_round_n)
        if has_shape_len(main, 1) and has_shape_len(baseline, 1):
            mean_leff = main/baseline
            return mean_leff, np.zeros(len(main))

        # Output of multiple experiments passed (nth_experiment, accuracy_at_round_n)
        elif has_shape_len(main, 2) and has_shape_len(baseline, 2):
            return __multi_experiment_leff(main, baseline)


        raise ValueError("Error in leff(). Expected main and baseline to have the same shape. Received shapes \"main={}\" and \"base={}\".".format(main.shape, baseline.shape))

    
    # Calculate labeling efficiency for single value pair (values of different functions at active learning step n)
    elif of_type(float, main, baseline) or of_type(int, main, baseline):
        return main/baseline
    

    raise ValueError("Error in leff(). Type missmatch. Expected parameters both to be of type np.ndarray, float or int.\
         Received types {} and {}.".format(type(main), type(baseline)))






import numpy as np


class ProblemUtils:

    def __init__(self, classification: bool=True, is_binary: bool=False):
        self.classification = classification

        self.is_binary = is_binary
        if not self.classification:
            self.is_binary = False


    def extend_binary_predictions(self, predictions, num_classes=2):
        """
            In MC Dropout case always predictions of shape
            (batch_size, sample_size, classes) for classification 
            or (batch_size, sample_self.is_binary = is_binarysize) for binary/regression case
        """

        # Don't modify predictions shape in regression case
        if not self.classification:
            return predictions


        # Binary case: calculate complementary prediction and concatenate
        if self.is_binary:
            bin_alt_class = (1 + np.zeros(predictions.shape)) - predictions

            # Expand dimensions for predictions to concatenate. Is this needed?
            bin_alt_class = np.expand_dims(bin_alt_class, axis=-1)
            predictions = np.expand_dims(predictions, axis=-1)

            # Concatenate predictions
            predictions = np.concatenate([predictions, bin_alt_class], axis=len(predictions.shape)-1)
        
        return predictions

    
    def alt_extend_binary_predictions(self, predictions):
        """
            Extend dimensions for binary classification problems.
 
            Parameters:
                predictions (numpy.ndarray): Predictions made by the network.

            Returns:
                (numpy.ndarray) The predictions with extended dimension.
        """
        # Don't modify predictions shape in regression case
        if not self.classification:
            return predictions

        # Binary case: calculate complementary prediction and concatenate
        if self.is_binary():
            bin_alt_class = (1 + np.zeros(predictions.shape)) - predictions

            # Expand dimensions for predictions to concatenate. Is this needed?
            # bin_alt_class = np.expand_dims(bin_alt_class, axis=-1)
            # predictions = np.expand_dims(predictions, axis=-1)

            # Concatenate predictions
            class_axis = len(predictions.shape) + 1
            predictions = np.concatenate([predictions, bin_alt_class], axis=len(predictions.shape)-1)
        
        return predictions


import numpy as np
from tensorflow.keras.metrics import get

class Loss:

    def __init__(self, name="loss"):
        self.name = name
        self.__fn = get(name)


    def __call__(self, true_targets, predictions, **kwargs):
        exp, _var = predictions
        return np.mean(self.__fn(true_targets, exp))


import numpy as np

class TruePositive:

    def __init__(self, multi_class=True, labels=None):
        self.name = "true_positive"
        self._multi_class = multi_class
        self._labels = labels

    
    def __call__(self, true_targets, predictions, **kwargs):
        """
        
        """
        pass


        # Binary case
        # predictions = np.where(predictions>.5, 1, 0)
        # stacked_targets = np.vstack([true_targets]*num_samples).T
        # predicted_labels = np.argmax(predictions, axis=-1)
        # return np.sum(stacked_targets == predicted_labels]



import numpy as np
import tensorflow as tf
from tensorflow.keras.metrics import get

class Accuracy:

    def __init__(self, name="accuracy"):
        self.name = name

    
    def __call__(self, true_targets, predictions, **kwargs):
        """
            Calculate the accuracy for the sampling procedure.
        """
        pred_mean = np.mean(predictions, axis=1)
        acc_fn = get(self.name)
        true_targets = tf.convert_to_tensor(true_targets)
        output = acc_fn(true_targets, pred_mean)
        return np.mean(output)

        # pred_targets = np.argmax(predictions, axis=-1)
        # if len(targets.shape) == 2:
        #     targets = np.argmax(targets, axis=1)

        # if sample_size > 1:
        #     targets = np.vstack([targets]*sample_size).T
        
        # return np.mean(pred_targets == targets)


    def __str__(self):
        return self.name


import numpy as np

"""
    A collection of helper functions to do check ups for score calculation.
"""

def has_shape_len(input_arr: np.ndarray, shape_length: int) -> bool:
    return len(input_arr.shape) == shape_length


def has_same_len(first_arr: np.ndarray, second_arr: np.ndarray) -> bool:
    return len(first_arr) == len(second_arr)


def of_type(_type, value_1, *args) -> bool:
    """
        Check if a collection of values are of the same type.

        Parameters:
            _type (any): The type to check for.
            value_1 (any): The first value to check.
            *args (any): Rest of values to check against given type.

        Returns:
            (bool) whether or not all inputs of given type. 
    """

    all_of_type = isinstance(value_1, _type)
    i = len(args)

    while i > 0 and all_of_type != False:
        all_of_type = isinstance(args[i-1], _type)
        i -= 1

    return all_of_type
    


from .. import Pool


def datapoints_in_use(pool: Pool) -> float:
    """
        Parameters:
            pool (Pool): The pool managing the labeled and unlabeled samples

        Returns:
            (float) percentage of used up total unlabeled pool size.
    """
    len_labeled = pool.get_length_labeled()
    len_unlabeled = pool.get_length_unlabeled()

    return len_labeled/(len_labeled+len_unlabeled)



from .false_positive import FalsePositive
from .false_negative import FalseNegative
from .true_negative import TrueNegative
from .true_positive import TruePositive
from .auc import AUC
from .accuracy import Accuracy
from .loss import Loss


import numpy as np

class TrueNegative:

    def __init__(self):
        self.name = "true_negative"
        self._multi_label = multi_label

    
    def __call__(self, true_targets, predictions, sample_size=None, **kwargs):
        num_samples = predictions.shape[1]
        stacked_targets = np.vstack([true_targets]*num_samples).T
        predicted_labels = np.argmax(predictions, axis=-1)

        (stacked_targets == predicted_labels)


import numpy as np
from .__checks import has_same_len, has_shape_len, of_type


def __single_experiment_qeff(main:np.ndarray, baseline: np.ndarray) -> tuple:
    main_mean = np.mean(main)
    main_std = np.std(main)

    base_mean = np.mean(baseline)
    base_std = np.std(baseline)
    return (main_mean-base_mean)/max(main_mean, base_mean), \
        (main_std-base_std)/max(main_std, base_std)


def __multi_experiment_qeff(main: np.ndarray, baseline: np.ndarray) -> tuple:
    main_exp_mean = np.mean(np.mean(main, axis=0))
    main_exp_std = np.mean(np.std(main, axis=0))

    base_exp_mean = np.mean(np.mean(baseline, axis=0))
    base_exp_std = np.mean(np.std(baseline, axis=0))

    return (main_exp_mean-base_exp_mean)/max(main_exp_mean, base_exp_mean), \
         (main_exp_std-base_exp_std)/max(main_exp_std, base_exp_std)


# def qeff(main, baseline, args*):
def qeff(main, baseline):
    """
        Calculate the query efficiency of main method to
        the baseline.

        Parameters:
            main (np.ndarray): The query time per timestep
            baseline (np.ndarray): The baseline times to check against


        Examples:
        > queff(1.3123, 5.32132)

        > queff(np.array([8.93912, 7.93938, 7.8848]), np.array([3.93912, 3.93938, 3.8848]))

        > mult_exp_1 = np.array([[1.313, 1.238, 1.134], [1.434, 1.313, 1.223]])
        > mult_exp_2 = np.array([[5.313, 4.238, 3.134], [5.434, 4.313, 3.223]])
        > queff(mult_exp_1, mult_exp_2)
    """

    if of_type(np.ndarray, main, baseline):

        if not has_same_len(main, baseline):
            raise ValueError("Error in qeff(). Expected length of input parameters to match.")
        
        # Output of single experiment was passed (query_time_at_al_round_n)
        if has_shape_len(main, 1) and has_shape_len(baseline, 1):
            return __single_experiment_qeff(main, baseline)

        # Output of multiple experiments was passed (n_th_experiment, query_time_at_al_round_n)
        elif has_shape_len(main, 2) and has_shape_len(main, 2):
            return __multi_experiment_qeff(main, baseline)

        raise ValueError("Error in qeff(). Invalid dimensionality. Inputs can't have more than 2 dimensions. Received shapes \"main={}\" and \"baseline={}\".".format(main.shape, baseline.shape))

    # Single value
    elif of_type(float, main, baseline) or of_type(int, main, baseline):
        return (main-baseline)/max(main, baseline)


    raise ValueError("Error in qeff(). Expected both inputs to be of type numpy.ndarray, float or int. Got \"main={}\" and \"baseline={}\"".format(type(main), type(baseline)))



import os, math
import uuid
import numpy as np
from enum import Enum
import tensorflow as tf

from . import Checkpoint
from ..utils import setup_logger, ProblemUtils
from ..stats import get


class Model:
    """
        Base wrapper for deep learning models to interface
        with the active learning environment. 
        
        Attributes:
            _model (tf.Model): Tensorflow or pytorch module.
            _config (Config): Model configuration
            _mode (Mode): The mode the model is in 'train' or 'test'/'eval'.
            _model_type (str): The model type
            _checkpoints (Checkpoint): Created checkpoints.

        Parameters:
            model (tf.Model): The tensorflow model to be used.
            config (Config): Configuration object for the model. (default=None)
            is_binary (bool): 
            classification (bool): 
    """

    def __init__(
        self, 
        model, 
        config=None, 
        name=None,
        model_type=None, 
        checkpoint=None,
        verbose=False,
        checkpoint_path=None,
        **kwargs
    ):

        self.__verbose = verbose
        self.logger = setup_logger(verbose, "Model Logger")
        self.__id = uuid.uuid1()
        self._model = model
        self._config = config
        self._model_type = model_type
        self._name = name
        self._compile_params = None
        self._problem = ProblemUtils(
            kwargs.get("classification", True),
            kwargs.get("is_binary", False)
        )

        self.eval_metrics = []

        # Checkpoints path set?
        if checkpoint_path is None:
            checkpoint_path = os.getcwd()
        self._checkpoints = Checkpoint(checkpoint_path) if checkpoint is None else checkpoint



    def __call__(self, inputs, **kwargs):
        return self._model(inputs, **kwargs)


    def predict(self, inputs, **kwargs):
        """
            Approximate predictive distribution.

            Parameter:
                inputs (numpy.ndarray): The inputs for the approximation

        """
        return self._model.predict(inputs, **kwargs)
    

    def evaluate(self, inputs, targets, **kwargs):
        """
            Evaluate a model on given input data and targets.

            Parameters:
                inputs (numpy.ndarray):
                targets (numpy.ndarray):

            Returns:
                (list) A list with two values. [loss, accuracy]  
        """

        pred_targets = self.predict(inputs, **kwargs)
        output_metrics = {}
        for metric in self.eval_metrics:
            
            if isinstance(metric, str):
                metric = tf.keras.metrics.get(metric)

            metric_name = None
            if hasattr(metric, "__name__"):
                metric_name = metric.__name__
            
            else:
                metric_name = metric.name

            output_metrics[metric_name] = metric(targets, pred_targets)
            
        return output_metrics


    def fit(self, *args, **kwargs):
        """
            Fit the model to the given data.

            Args:
                x (numpy.ndarray): The inputs to train the model on. (default=None)
                y (numpy.ndarray): The targets to fit the model to. (default=None)
                batch_size (int): The size of each individual batch

            Returns:
                () a record of the trianing procedure
        """

        if self._config is not None and "fit" in self._config and isinstance(self._config["fit"], dict):
            fit_params = self._config["fit"]
            kwargs.update(fit_params)

        return self._model.fit(*args, **kwargs)


    def compile(self, *args, **kwargs):
        """
            Compile the model if needed
        """
        self._compile_params = kwargs
        self._model.compile(**kwargs)

        # Create evaluation metrics from compile parameters
        metrics = self._create_init_metrics(kwargs)
        if self.eval_metrics == []:
            self.eval_metrics = metrics


    def disable_batch_norm(self):
        """
            Disable batch normalization for activation of dropout during prediction.

            Parameters:
                - model (tf.Model) Tensorflow neural network model.
        """
        
        disabled = False
        for l in self._model.layers:
            if l.__class__.__name__ == "BatchNormalization":
                disabled = True
                l.trainable = False

        if disabled:
            self.logger.info("Disabled BatchNorm-Layers.")


    def clear_session(self):
        tf.keras.backend.clear_session()



    # ------
    # Model runtime configurations
    # -----------------------------

    def reset(self, pool, dataset):
        """
            Use to reset states, weights and other stuff after each active learning loop iteration.

            Parameters:
                pool (Pool): The pool managing labeled and unlabeled indices.
                dataset (Dataset): The dataset containting the different splits.
        """
        self._model = tf.keras.models.clone_model(self._model)
        self._model.compile(**self._compile_params)


    def optimize(self, inputs, targets):
        """
            Use to perform optimization during active learning loop.
        """
        pass


    # ----------------------
    # Utilities
    # ------------------------------

    def batch_prediction(self, inputs, batch_size=1, **kwargs):
        """
            
            Parameters:
                inputs (numpy.ndarray): Inputs going into the model
                n_times (int): How many times to sample from posterior?
                batch_size (int): In how many batches to split the data?
        """
    
        if batch_size < 1:
            raise ValueError("Error in McDropout.__call__(). Can't select negative amount of batches.")

        total_len = len(inputs)
        num_batches = math.ceil(total_len/batch_size)
        batches = np.array_split(inputs, num_batches, axis=0)

        predictions = []
        for batch in batches:
            predictions.append(self._model(batch, training=True))

        return np.vstack(predictions)


    def _create_init_metrics(self, kwargs):
        """
            Extract loss and other metrics passed to Model.compile()
            and transfer them to metrics which should be used for evaluation.
        """

        loss = kwargs.get("loss", None)
        metrics = kwargs.get("metrics", [])

        if not isinstance(metrics, list):
            metrics = [metrics]

        if loss is not None:
            metrics = [loss] + metrics
        
        return metrics


    def _extract_metric_names(self, metrics):
        """
            Get the metric names from a list of metric classes, 
            passing already existing metric names on.

            Parameters:
                metrics (list()): A list can contain strings and metric objects.
            
            Returns:
                (list(str)) a list of metric names
        """
        
        all_metrics = []
        for metric in metrics:

            metric_name = metric
            if not isinstance(metric, str):
                metric_name = metric.name
            
            all_metrics.append(metric_name)
        
        return all_metrics
        

    def _init_metrics(self, prefix, metrics):
        """
            Initializes metrics passed to the object during compilation.
            Using model specific metrics instead. 

            Parameters:
                prefix (str): The sub-package of tf_al.stats to be used (sampling, stochastic, ...)
                metrics (list(str)): A list of metric names.
            
            Returns:
                (list) of initialized metrics
        """
        initialized_metrics = []
        for metric in metrics:
            initialized_metrics.append(get(prefix, metric))

        return initialized_metrics




    # --------------
    # Checkpoint creation/loading
    # ------------------------------

    def empty_checkpoint(self):
        return self._checkpoints.empty()

    def new_checkpoint(self):
        self._checkpoints.new(self._model)

    
    def load_checkpoint(self, iteration=None):
        self._checkpoints.load(self._model, iteration)


    def clear_checkpoints(self):
        self._checkpoints.clean()


    def save_weights(self):
        path = self._checkpoints.PATH
        self._model.save_weights(path)


    def load_weights(self):
        path = self._checkpoints.PATH
        self._model.load_weights(path)
        

    def has_save_state(self):
        try:
            self.load_weights()
            return True

        except:
            return False



    # -----------
    # Access Configuration
    # ----------------------------

    def get_config_for(self, name):
        if (self._config is not None) and name in self._config:
            return self._config[name]
        
        return {}


    def get_config(self):
        return self._config.kwargs


    # --------------
    # Access important flags for predictions
    # ----------------------------

    def is_classification(self):
        return self._problem.classification


    def is_binary(self):
        return self._problem.is_binary


    # ---------------
    # Acquisition functions
    # --------------------------

    def get_query_fn(self, name):
        """
            Get model specific acquisition function.

            Parameters:
                name (str): The name of the acquisition function to return.

            Returns:
                (function) the acquisition function to use.
        """
        pass


    def __max_entropy(self, data, **kwargs):
        pass

    def __bald(self, data, **kwargs):
        pass

    def __max_var_ratio(self, data, **kwargs):
        pass

    def __std_mean(self, data, **kwargs):
        pass
    

    # -------------
    # Metric hooks
    # -----------------

    # def _on_evaluate_loss(self, **kwargs):
    #     pass

    
    # def _on_evaluate_acc(self, **kwargs):
    #     pass


    # -----------------
    # Setter/-Getter
    # --------------------------

    def get_id(self):
        return str(self.__id)

    def get_model_name(self, prefix=True):
        """
            Returns the model name.

            Parameters:
                prefix (bool): Prefix the model name with model type?

            Returns:
                (str) the model name.
        """

        # Model name and model type?
        model_name_exists = self._name is not None
        if self._model_type is not None:
            if model_name_exists and prefix:
                return self._model_type.lower() + "_" + self._name

            return self._model_type.lower()
        
        # Only model name given
        if model_name_exists:
            return self._name

        # Default model name
        return "model"
        

    def get_model(self):
        return self._model

    def get_metric_names(self):
        return self._model.metrics_names

    def get_base_model(self):
        return self._model
        

    # ---------------
    # Dunder
    # ----------------------
    
    def __str__(self):
        return self.get_model_name()

    def __getstate__(self):
        """
            Called when pickeling the object.
            Remove logger, because no use to pickle logger.
        """
        d = dict(self.__dict__)
        del d["logger"]
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        verbose = d["__verbose"]
        self.logger = setup_logger(verbose, "Model Logger")



from scipy.special import comb
from sklearn.preprocessing import label_binarize
import numpy as np


class AUC():

    def __init__(self, num_thresholds=10, multi_label=False, from_logits=False, labels=None):
        self.name = "auc"

        if num_thresholds < 1:
            raise ValueError("Error in AUC(). Number of thresholds needs to be > 1.")
        self._thresholds = np.linspace(0, 1, num_thresholds)
        self._from_logits = from_logits
        self._multi_label = multi_label
        self._labels = labels

    
    def __call__(self, true_target, predictions, sample_size=None):
        """
            Calculate the AUC metric.
        """
        # https://towardsdatascience.com/roc-and-auc-how-to-evaluate-machine-learning-models-in-no-time-fb2304c83a7f
        # https://towardsdatascience.com/confusion-matrix-for-your-multi-class-machine-learning-model-ff9aa3bf7826#:~:text=False%20Positive%20(FP)%3A%20It,the%20negative%20class%20as%20positive.&text=It%20is%20also%20known%20as,or%20(1%2DAccuracy).

        if self._multi_label:
            return self.__multiclass(true_target, predictions)

                        

        return 12


    def __multiclass(self, true_target, predictions):

        
        num_classes = predictions.shape[-1]
        class_labels = self._get_labels(num_classes)
        true_target = label_binarize(true_target)
        num_thresholds = len(self._thresholds)

        shape = (num_classes, num_thresholds)
        tp = np.zeros(shape)
        tn = np.zeros(shape)
        fp = np.zeros(shape)
        fn = np.zeros(shape)

        for idx in range(num_classes):
            label = class_labels[idx]

            class_prediction = np.take(predictions, idx, axis=-1)
            for th_idx in range(num_thresholds):
                
                threshold = self._thresholds[th_idx]
                selector = true_target == label

                current_tp = np.sum(true_target[selector] == pred_targets[selector])


                other_selector = np.logical_not(selector)
                # current_fn = np.sum(true_target[other_selector] == )
                tp[idx, th_idx] = current_tp



    def true_positive(self, true_targets, predictions):

        pred_targets = np.where(np.mean(class_prediction, axis=1)>=threshold, label, -1)

    
    def __binary(self, true_target, predictions):

        pass



    def _get_labels(self, num_classes):
        if self._labels is not None:
            return self._labels

        return list(range(num_classes))


    